
__all__ = [
    "APIError",
    "fatal_exception",   
    "LagoApiError",
    "ReportingApiError",
]

class APIError(Exception):
    def __init__(self, url: str, status, payload):
        self.url = url
        self.status = status
        self.payload = payload

    def __str__(self):
        return f"[Lago] url: {self.url}, status: {self.status}, payload: {self.payload}"


def fatal_exception(exc):
    if isinstance(exc, APIError):
        # retry on server errors and client errors
        # with 429 status code (rate limited),
        # don't retry on other client errors
        return (400 <= exc.status < 500) and exc.status != 429
    else:
        # retry on all other errors (eg. network)
        return False


class LagoApiError(Exception):
    ...

class ReportingApiError(Exception):
    ...



from typing import Optional, Type, Dict, Union, List, Type, Any
from aiolago.types.base import BaseRoute, BaseResource, lazyproperty
from aiolago.types.errors import LagoApiError
from aiolago.schemas.customer import CustomerUsageResponse

__all__ = [
    'Subscription',
    'SubscriptionResponse',
    'SubscriptionRoute',
    'SubscriptionUsageResponse',
]


class Subscription(BaseResource):
    plan_code: Optional[str]
    external_customer_id: Optional[str]
    name: Optional[str]
    external_id: Optional[str]
    subscription_date: Optional[str]
    billing_time: Optional[str]


class SubscriptionResponse(BaseResource):
    lago_id: str
    lago_customer_id: Optional[str]
    external_customer_id: Optional[str]
    canceled_at: Optional[str]
    created_at: Optional[str]
    plan_code: Optional[str]
    started_at: Optional[str]
    status: Optional[str]
    name: Optional[str]
    external_id: Optional[str]
    billing_time: Optional[str]
    terminated_at: Optional[str]
    subscription_date: Optional[str]
    previous_plan_code: Optional[str]
    next_plan_code: Optional[str]
    downgrade_plan_date: Optional[str]


class SubscriptionUsageResponse(SubscriptionResponse):
    usage: Optional[CustomerUsageResponse]


class SubscriptionRoute(BaseRoute):
    input_model: Optional[Type[BaseResource]] = Subscription
    response_model: Optional[Type[BaseResource]] = SubscriptionResponse
    usage_model: Optional[Type[BaseResource]] = None

    @lazyproperty
    def api_resource(self):
        return 'subscriptions'

    @lazyproperty
    def root_name(self):
        return 'subscription'
    
    def find(
        self, 
        external_customer_id: str,
        resource_id: Optional[str] = None,
        external_subscription_id: Optional[str] = None,
        params: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> SubscriptionUsageResponse:
        """
        GET a Single Subscription

        :param resource_id: The ID of the Resource to GET [or external_subscription_id]
        :param external_subscription_id: The ID of the External Subscription [or resource_id]
        :param external_customer_id: The ID of the External Customer
        :param params: Optional Query Parameters
        """
        # So this is a bit different since we need to fetch all subscriptions first
        # and then fetch the usage information
        subscriptions: List[SubscriptionResponse] = self.find_all(
            external_customer_id = external_customer_id,
            params = params,
            with_index = False,
        )
        external_subscription_id = external_subscription_id or resource_id
        subscription: SubscriptionResponse = next(
            (
                sub
                for sub in subscriptions
                if (sub.external_id == external_subscription_id or sub.lago_id == external_subscription_id)
            ),
            None,
        )
        if subscription is None:
            raise LagoApiError(f'Could not find subscription with external_customer_id: {external_subscription_id}')

        api_resource = f'customers/{subscription.external_customer_id}/current_usage'
        if not params: params = {}
        params['external_subscription_id'] = subscription.external_id
        api_response = self._send(
            method = 'GET',
            url = api_resource, 
            params = params,
            headers = self.headers,
            **kwargs
        )
        data = self.handle_response(api_response)
        if data: data = data.json().get("customer_usage")
        usage = self.prepare_response(data, response_object = CustomerUsageResponse)
        return SubscriptionUsageResponse(**subscription.dict(), usage = usage)


    async def async_find(
        self, 
        external_customer_id: str,
        resource_id: Optional[str] = None,
        external_subscription_id: Optional[str] = None,
        params: Optional[Dict[str, Any]] = None,
        **kwargs
    )  -> SubscriptionUsageResponse:
        """
        GET a Single Subscription

        :param resource_id: The ID of the Resource to GET [or external_subscription_id]
        :param external_subscription_id: The ID of the External Subscription [or resource_id]
        :param external_customer_id: The ID of the External Customer
        :param params: Optional Query Parameters
        """
        subscriptions: List[SubscriptionResponse] = await self.async_find_all(
            external_customer_id = external_customer_id,
            params = params,
            with_index = False,
        )
        external_subscription_id = external_subscription_id or resource_id
        subscription: SubscriptionResponse = next(
            (
                sub
                for sub in subscriptions
                if (sub.external_id == external_subscription_id or sub.lago_id == external_subscription_id)
            ),
            None,
        )
        if subscription is None:
            raise LagoApiError(f'Could not find subscription with external_customer_id: {external_subscription_id}')

        api_resource = f'customers/{subscription.external_customer_id}/current_usage'
        if not params: params = {}
        params['external_subscription_id'] = subscription.external_id
        api_response = await self._async_send(
            method = 'GET',
            url = api_resource, 
            params = params,
            headers = self.headers,
            **kwargs
        )
        data = self.handle_response(api_response)
        if data: data = data.json().get("customer_usage")
        usage = self.prepare_response(data, response_object = CustomerUsageResponse)
        return SubscriptionUsageResponse(**subscription.dict(), usage = usage)

    
    def get(
        self, 
        external_customer_id: str,
        resource_id: Optional[str] = None,
        external_subscription_id: Optional[str] = None,
        params: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> SubscriptionUsageResponse:
        """
        GET a Single Subscription

        :param resource_id: The ID of the Resource to GET [or external_subscription_id]
        :param external_subscription_id: The ID of the External Subscription [or resource_id]
        :param external_customer_id: The ID of the External Customer
        :param params: Optional Query Parameters
        """
        return self.find(external_subscription_id = external_subscription_id, external_customer_id = external_customer_id, resource_id = resource_id, params = params, **kwargs)
    
    async def async_get(
        self,
        external_customer_id: str,
        resource_id: Optional[str] = None,
        external_subscription_id: Optional[str] = None,
        params: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> SubscriptionUsageResponse:
        """
        GET a Single Subscription

        :param resource_id: The ID of the Resource to GET [or external_subscription_id]
        :param external_subscription_id: The ID of the External Subscription [or resource_id]
        :param external_customer_id: The ID of the External Customer
        :param params: Optional Query Parameters
        """
        return await self.async_find(external_subscription_id = external_subscription_id, external_customer_id = external_customer_id, resource_id = resource_id, params = params, **kwargs)

    
    def find_all(
        self, 
        resource_id: Optional[str] = None,
        external_customer_id: Optional[str] = None,
        params: Optional[Dict[str, Any]] = None,
        with_index: Optional[bool] = False,
        **kwargs
    ) -> Union[Dict[str, Union[List[SubscriptionResponse], Dict[str, Any]]], List[SubscriptionResponse]]:
        """
        GET all available subscriptions for a customer

        :param resource_id: The ID of the Resource to GET [or external_customer_
        :param external_customer_id: The customer ID
        :param params: Optional Query Parameters
        :param with_index: If set to true, the index of the resource will be returned

        :return: Dict[str, Union[List[Type[SubscriptionResponse]], Dict[str, Any]]]
        """
        if params is None: params = {}
        params['external_customer_id'] = external_customer_id or resource_id
        api_response = self._send(
            method = 'GET',
            url = self.api_resource,
            params = params,
            headers = self.headers,
            timeout = self.timeout,
            **kwargs
        )
        data = self.handle_response(api_response)
        if data: data = data.json()
        result = self.prepare_index_response(data)
        if not with_index: result = result.get(self.api_resource)
        return result

    async def async_find_all(
        self, 
        resource_id: Optional[str] = None,
        external_customer_id: Optional[str] = None,
        params: Optional[Dict[str, Any]] = None,
        with_index: Optional[bool] = False,
        **kwargs
    ) -> Union[Dict[str, Union[List[SubscriptionResponse], Dict[str, Any]]], List[SubscriptionResponse]]:
        """
        GET all available subscriptions for a customer

        :param resource_id: The ID of the Resource to GET [or external_customer_
        :param external_customer_id: The customer ID
        :param params: Optional Query Parameters
        :param with_index: If set to true, the index of the resource will be returned

        :return: Dict[str, Union[List[Type[SubscriptionResponse]], Dict[str, Any]]]
        """
        if params is None: params = {}
        params['external_customer_id'] = external_customer_id or resource_id
        api_response = await self._async_send(
            method = 'GET',
            url = self.api_resource,
            params = params,
            headers = self.headers,
            timeout = self.timeout,
            **kwargs
        )
        data = self.handle_response(api_response)
        if data: data = data.json()
        result = self.prepare_index_response(data)
        if not with_index: result = result.get(self.api_resource)
        return result
    
    def get_all(
        self,
        resource_id: Optional[str] = None,
        external_customer_id: Optional[str] = None,
        params: Optional[Dict[str, Any]] = None,
        with_index: Optional[bool] = False,
        **kwargs
    ) -> Union[Dict[str, Union[List[SubscriptionResponse], Dict[str, Any]]], List[SubscriptionResponse]]:
        """
        GET all available subscriptions for a customer

        :param resource_id: The ID of the Resource to GET [or external_customer_
        :param external_customer_id: The customer ID
        :param params: Optional Query Parameters
        :param with_index: If set to true, the index of the resource will be returned

        :return: Dict[str, Union[List[Type[SubscriptionResponse]], Dict[str, Any]]]
        """
        return self.find_all(resource_id = resource_id, external_customer_id = external_customer_id, params = params, with_index = with_index, **kwargs)
    
    async def async_get_all(
        self,
        resource_id: Optional[str] = None,
        external_customer_id: Optional[str] = None,
        params: Optional[Dict[str, Any]] = None,
        with_index: Optional[bool] = False,
        **kwargs
    ) -> Union[Dict[str, Union[List[SubscriptionResponse], Dict[str, Any]]], List[SubscriptionResponse]]:
        """
        GET all available subscriptions for a customer

        :param resource_id: The ID of the Resource to GET [or external_customer_
        :param external_customer_id: The customer ID
        :param params: Optional Query Parameters
        :param with_index: If set to true, the index of the resource will be returned

        :return: Dict[str, Union[List[Type[SubscriptionResponse]], Dict[str, Any]]]
        """
        return await self.async_find_all(resource_id = resource_id, external_customer_id = external_customer_id, params = params, with_index = with_index, **kwargs)
    

    """
    Extra Methods
    """

    def exists(
        self,
        external_customer_id: str,
        resource_id: Optional[str] = None,
        external_subscription_id: Optional[str] = None,
        params: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> Union[SubscriptionUsageResponse, bool]:
        """
        See whether a Subscription Exists

        :param resource_id: The ID of the Resource to GET [or external_subscription_id]
        :param external_subscription_id: The ID of the External Subscription [or resource_id]
        :param external_customer_id: The ID of the External Customer
        :param params: Optional Query Parameters
        """
        try:
            return self.find(external_customer_id = external_customer_id, resource_id = resource_id, external_subscription_id = external_subscription_id, params = params,  **kwargs)
        except Exception:
            return False
    
    async def async_exists(
        self,
        external_customer_id: str,
        resource_id: Optional[str] = None,
        external_subscription_id: Optional[str] = None,
        params: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> Union[SubscriptionUsageResponse, bool]:
        """
        See whether a Resource Exists

        :param resource_id: The ID of the Resource to GET [or external_subscription_id]
        :param external_subscription_id: The ID of the External Subscription [or resource_id]
        :param external_customer_id: The ID of the External Customer
        :param params: Optional Query Parameters
        """
        try:
            return await self.async_find(external_customer_id = external_customer_id, resource_id = resource_id, external_subscription_id = external_subscription_id, params = params, **kwargs)
        except Exception:
            return False
    
    def upsert(
        self,
        external_customer_id: str,
        resource_id: Optional[str] = None,
        external_subscription_id: Optional[str] = None,
        input_object: Optional[Type[Subscription]] = None,
        update_existing: bool = False, 
        overwrite_existing: bool = False,
        **kwargs
    ):
        """
        Upsert a Subscription
        Validates whether the Subscription exists, and if it does, updates it. 
        If it doesn't, creates it.
        If update_existing is True, it will always update the Resource
        If overwrite_existing is True, it will re-create the Resource

        :resource_id: The ID of the Resource to Upsert
        :param external_subscription_id: The ID of the External Subscription [or resource_id]
        :param external_customer_id: The ID of the External Customer
        :param input_object: Input Object to Upsert
        :param update_existing (bool): Whether to update the Resource if it exists
        :overwrite_existing (bool): Whether to overwrite the Resource if it exists
        """
        resource = self.exists(external_customer_id = external_customer_id, resource_id = resource_id, external_subscription_id = external_subscription_id, **kwargs)
        if resource is not None:
            if update_existing:
                return self.update(input_object = input_object, identifier = (external_subscription_id or resource_id), **kwargs)
            if overwrite_existing:
                self.destroy(resource_id = (external_subscription_id or resource_id), **kwargs)
                return self.create(
                    input_object = input_object, 
                    external_id = kwargs.pop('external_id', (external_subscription_id or resource_id)),
                    external_customer_id = external_customer_id,
                    **kwargs
                )
            return resource
        return self.create(
            input_object = input_object, 
            external_id = kwargs.pop('external_id', (external_subscription_id or resource_id)),
            external_customer_id = external_customer_id,
            **kwargs
        )
    
    async def async_upsert(
        self,
        external_customer_id: str,
        resource_id: Optional[str] = None,
        external_subscription_id: Optional[str] = None,
        input_object: Optional[Type[Subscription]] = None,
        update_existing: bool = False, 
        overwrite_existing: bool = False,
        **kwargs
    ):
        """
        Upsert a Resource
        Validates whether the Resource exists, and if it does, updates it. 
        If it doesn't, creates it.
        If update_existing is True, it will always update the Resource
        If overwrite_existing is True, it will re-create the Resource

        :param external_subscription_id: The ID of the External Subscription [or resource_id]
        :param external_customer_id: The ID of the External Customer
        :resource_id: The ID of the Resource to Upsert
        :param input_object: Input Object to Upsert
        :param update_existing (bool): Whether to update the Resource if it exists
        :overwrite_existing (bool): Whether to overwrite the Resource if it exists
        """
        resource = await self.async_exists(external_customer_id = external_customer_id, resource_id = resource_id, external_subscription_id = external_subscription_id, **kwargs)
        if resource is not None:
            if update_existing:
                return self.async_update(input_object = input_object, identifier = (external_subscription_id or resource_id), **kwargs)
            if overwrite_existing:
                await self.async_destroy(resource_id = (external_subscription_id or resource_id), **kwargs)
                return await self.async_create(
                    input_object = input_object, 
                    external_id = kwargs.pop('external_id', (external_subscription_id or resource_id)),
                    external_customer_id = external_customer_id,
                    **kwargs
                )
            return resource
        return await self.async_create(
            input_object = input_object, 
            external_id = kwargs.pop('external_id', (external_subscription_id or resource_id)),
            external_customer_id = external_customer_id,
            **kwargs
        )





import inspect
from datetime import datetime, timedelta

__all__ = [
    'is_naive',
    'total_seconds',
    'remove_trailing_slash',
    'full_name',
]

def full_name(func, follow_wrapper_chains=True):
    """
    Return full name of `func` by adding the module and function name.

    If this function is decorated, attempt to unwrap it till the original function to use that
    function name by setting `follow_wrapper_chains` to True.
    """
    if follow_wrapper_chains: func = inspect.unwrap(func)
    return f'{func.__module__}.{func.__qualname__}'

def is_naive(dt: datetime):
    """Determines if a given datetime.datetime is naive."""
    return dt.tzinfo is None or dt.tzinfo.utcoffset(dt) is None


def total_seconds(delta: timedelta):
    """Determines total seconds with python < 2.7 compat."""
    # http://stackoverflow.com/questions/3694835/python-2-6-5-divide-timedelta-with-timedelta
    return (delta.microseconds + (delta.seconds + delta.days * 24 * 3600) * 1e6) / 1e6


def remove_trailing_slash(host: str):
    return host[:-1] if host.endswith("/") else host




## Base Object Models
from aiolago.schemas.applied_add_on import AppliedAddOn, AppliedAddOnResponse
from aiolago.schemas.applied_coupon import AppliedCoupon, AppliedCouponResponse
from aiolago.schemas.billable_metric import BillableMetric, BillableMetricResponse, MetricGroup, MetricGroupResponse
from aiolago.schemas.coupon import Coupon, CouponResponse
from aiolago.schemas.plan import Plan, Charges, Charge, PlanResponse
from aiolago.schemas.add_on import AddOn, AddOnResponse
from aiolago.schemas.organization import Organization, OrganizationResponse
from aiolago.schemas.event import Event, BatchEvent, EventResponse
from aiolago.schemas.customer import Metric, ChargeObject, ChargeUsage, CustomerUsageResponse, Customer, BillingConfiguration, CustomerResponse
from aiolago.schemas.invoice import InvoiceStatusChange, InvoiceResponse
from aiolago.schemas.subscription import Subscription, SubscriptionResponse, SubscriptionUsageResponse
from aiolago.schemas.wallet import Wallet, WalletResponse
from aiolago.schemas.wallet_transaction import WalletTransaction, WalletTransactionResponse

## Route Models

from aiolago.schemas.applied_add_on import AppliedAddOnRoute
from aiolago.schemas.applied_coupon import AppliedCouponRoute
from aiolago.schemas.billable_metric import BillableMetricRoute
from aiolago.schemas.coupon import CouponRoute
from aiolago.schemas.plan import PlanRoute
from aiolago.schemas.add_on import AddOnRoute
from aiolago.schemas.organization import OrganizationRoute
from aiolago.schemas.event import EventRoute
from aiolago.schemas.customer import CustomerRoute
from aiolago.schemas.invoice import InvoiceRoute
from aiolago.schemas.subscription import SubscriptionRoute
from aiolago.schemas.wallet import WalletRoute
from aiolago.schemas.wallet_transaction import WalletTransactionRoute
from aiolago.schemas.webhooks import WebhookRoute


import base64
from typing import Optional, Type
from aiolago.types.base import BaseRoute, BaseModel, lazyproperty

__all__ = [
    'WebhookRoute',
]

class WebhookRoute(BaseRoute):

    @lazyproperty
    def public_key(self):
        api_response = self.client.get(
            url = 'webhooks/public_key',
            headers = self.headers
        )
        coded_response = self.handle_response(api_response).text
        return base64.b64decode(coded_response)



from typing import Optional, Type
from aiolago.types.base import BaseRoute, BaseResource, lazyproperty

__all__ = [
    'Coupon',
    'CouponResponse',
    'CouponRoute',
]

class Coupon(BaseResource):
    name: Optional[str]
    code: Optional[str]
    amount_cents: Optional[int]
    amount_currency: Optional[str]
    expiration: Optional[str]
    expiration_date: Optional[str]
    percentage_rate: Optional[float]
    coupon_type: Optional[str]
    frequency: Optional[str]
    frequency_duration: Optional[int]


class CouponResponse(BaseResource):
    lago_id: str
    name: str
    code: str
    amount_cents: int
    amount_currency: str
    created_at: str
    expiration: str
    expiration_date: Optional[str]
    percentage_rate: Optional[float]
    coupon_type: Optional[str]
    frequency: Optional[str]
    frequency_duration: Optional[int]


class CouponRoute(BaseRoute):
    input_model: Optional[Type[BaseResource]] = Coupon
    response_model: Optional[Type[BaseResource]] = CouponResponse
    usage_model: Optional[Type[BaseResource]] = None

    @lazyproperty
    def api_resource(self):
        return 'coupons'

    @lazyproperty
    def root_name(self):
        return 'coupon'



from typing import Optional, Type
from aiolago.types.base import BaseRoute, BaseResource, lazyproperty

__all__ = [
    'AppliedAddOn',
    'AppliedAddOnResponse',
    'AppliedAddOnRoute',
]

class AppliedAddOn(BaseResource):
    external_customer_id: str
    add_on_code: str
    amount_cents: Optional[int]
    amount_currency: Optional[str]


class AppliedAddOnResponse(BaseResource):
    lago_id: str
    lago_add_on_id: str
    add_on_code: str
    external_customer_id: str
    lago_customer_id: str
    amount_cents: int
    amount_currency: str
    created_at: str

class AppliedAddOnRoute(BaseRoute):
    input_model: Optional[Type[BaseResource]] = AppliedAddOn
    response_model: Optional[Type[BaseResource]] = AppliedAddOnResponse
    usage_model: Optional[Type[BaseResource]] = None

    @lazyproperty
    def api_resource(self):
        return 'applied_add_ons'

    @lazyproperty
    def root_name(self):
        return 'applied_add_on'
    


from typing import Optional, Type
from aiolago.types.base import BaseRoute, BaseResource, lazyproperty

__all__ = [
    'AppliedCoupon',
    'AppliedCouponResponse',
    'AppliedCouponRoute',
]

class AppliedCoupon(BaseResource):
    external_customer_id: str
    coupon_code: str
    amount_cents: Optional[int]
    amount_currency: Optional[str]
    percentage_rate: Optional[float]
    frequency: Optional[str]
    frequency_duration: Optional[int]


class AppliedCouponResponse(BaseResource):
    lago_id: str
    lago_coupon_id: str
    coupon_code: str
    external_customer_id: str
    lago_customer_id: str
    amount_cents: int
    amount_currency: str
    expiration_date: Optional[str]
    created_at: str
    terminated_at: Optional[str]
    percentage_rate: Optional[float]
    frequency: Optional[str]
    frequency_duration: Optional[int]

class AppliedCouponRoute(BaseRoute):
    input_model: Optional[Type[BaseResource]] = AppliedCoupon
    response_model: Optional[Type[BaseResource]] = AppliedCouponResponse
    usage_model: Optional[Type[BaseResource]] = None

    @lazyproperty
    def api_resource(self):
        return 'applied_coupons'

    @lazyproperty
    def root_name(self):
        return 'applied_coupon'


import aiohttpx
import datetime
import uuid
import functools

from typing import Optional, Callable, Dict, Union, List

from lazyops.types import lazyproperty
from lazyops.utils import is_coro_func, timer
from aiolago.schemas import *
from aiolago.utils.logs import logger
from aiolago.utils.config import LagoSettings
from aiolago.utils.config import settings as lago_settings

from aiolago.utils.helpers import full_name
from aiolago.routes import ApiRoutes


class LagoClient:
    """
    Main Client for all the routes in the API.
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        url: Optional[str] = None,
        scheme: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
        api_path: Optional[str] = None,
        headers: Optional[Dict] = None,
        debug_enabled: Optional[bool] = None,
        on_error: Optional[Callable] = None,
        timeout: Optional[int] = None,
        api_key_header: Optional[str] = None,
        ignore_errors: Optional[bool] = None,
        settings: Optional[LagoSettings] = None,
        **kwargs
    ):
        settings = settings or lago_settings

        self.api_key = api_key if api_key is not None else settings.api_key
        self.api_url = settings.get_api_url(host = host, port = port, scheme = scheme, url = url)
        self.base_url = settings.get_base_api_url(host = host, port = port, scheme = scheme, url = url, api_path = api_path)
        self.debug_enabled = debug_enabled if debug_enabled is not None else lago_settings.debug_enabled
        
        self.timeout = timeout if timeout is not None else settings.timeout
        self.headers = headers if headers is not None else settings.get_headers(api_key = self.api_key, api_key_header = api_key_header)
        self.on_error = on_error
        self.ignore_errors = ignore_errors if ignore_errors is not None else settings.ignore_errors
        self._kwargs = kwargs
        self.log_method = logger.info if self.debug_enabled else logger.debug
        self.client = aiohttpx.Client(
            base_url = self.base_url,
            timeout = self.timeout,
        )

        self.routes = ApiRoutes(
            client = self.client,
            headers = self.headers,
            debug_enabled = self.debug_enabled,
            on_error = self.on_error,
            ignore_errors = self.ignore_errors,
            timeout = self.timeout,
            **self._kwargs
        )
        logger.info(f"Lago Client initialized: {self.client.base_url}")
        if self.debug_enabled:
            logger.debug(f"Debug Enabled: {self.debug_enabled}")


    @lazyproperty
    def add_ons(self) -> AddOnRoute:
        return self.routes.add_ons
    
    @lazyproperty
    def applied_add_ons(self) -> AppliedAddOnRoute:
        return self.routes.applied_add_ons
    
    @lazyproperty
    def applied_coupons(self) -> AppliedCouponRoute:
        return self.routes.applied_coupons
    
    @lazyproperty
    def billable_metrics(self) -> BillableMetricRoute:
        return self.routes.billable_metrics
    
    @lazyproperty
    def coupons(self) -> CouponRoute:
        return self.routes.coupons
    
    @lazyproperty
    def customers(self) -> CustomerRoute:
        return self.routes.customers
    
    @lazyproperty
    def events(self) -> EventRoute:
        return self.routes.events
    
    @lazyproperty
    def invoices(self) -> InvoiceRoute:
        return self.routes.invoices
    
    @lazyproperty
    def organizations(self) -> OrganizationRoute:
        return self.routes.organizations
    
    @lazyproperty
    def plans(self) -> PlanRoute:
        return self.routes.plans
    
    @lazyproperty
    def subscriptions(self) -> SubscriptionRoute:
        return self.routes.subscriptions
    
    @lazyproperty
    def wallets(self) -> WalletRoute:
        return self.routes.wallets
    
    @lazyproperty
    def wallet_transactions(self) -> WalletTransactionRoute:
        return self.routes.wallet_transactions
    
    @lazyproperty
    def webhooks(self) -> WebhookRoute:
        return self.routes.webhooks
    

    """
    Context Managers
    """

    async def async_close(self):
        await self.client.aclose()
    
    def close(self):
        self.client.close()
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        self.close()
    
    async def __aenter__(self):
        return self
    
    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.async_close()


    """
    Function Decorators
    """
    def on_event(
        self,
        customer_id: Optional[Union[str, Callable]] = None,
        metric_id: Optional[Union[str, Callable]] = None,
        transaction_id: Optional[Union[str, Callable]] = None,
        
        properties: Optional[Union[Dict, Callable]] = None,
        timestamp: Optional[float] = None,
        exclude_properties: Optional[List[str]] = None,
        exclude_on: Optional[Callable] = None,
        extra_properties: Optional[Dict] = None,
        extra_properties_func: Optional[Callable] = None,
        include_duration: Optional[bool] = False,

        include_results: Optional[bool] = True,
        include_args_in_properties: Optional[bool] = False,
        include_kwargs_in_properties: Optional[bool] = False,
        usage_callback: Optional[Callable] = None,
        **kwargs,
    ):
        """
        Lago Event Decorator to log events to the Lago API.
        Most params are either _str_ or _Callable_ which will pass the function's arguments 
        and expects a _str_ to be returned. If required parameters are not provided, the 
        event logging will not occur.

        Example:

        >>> @Lago.on_event(customer_id = get_customer_id)
        ... def my_func(headers: Dict, params: Dict):
        ...     ...

        >>> @Lago.on_event(customer_id="1234")
        ... async def async_run_func():
        ...     ...

        >>> @Lago.on_event(customer_id="1234", metric_id="my_event")
        ... def run_func():
        ...     ...

        :param customer_id: The customer_id to log the event to. (required)
        :param metric_id: The `metric_id` which correlates to a `BillableMetrics` to log. [also known as `code`]. defaults to the function name.
        :param transaction_id: Unique identifer of the event in your application. Defaults to a unique uuid.

        :param properties: A dictionary of parameters to log with the event. Defaults 
        to the function's arguments.
        :param created_at: The time the event occurred. Defaults to the current time.

        :param exclude_properties: A list of properties to exclude from the event.
        :param exclude_on: A function that takes validates whether this event should be logged.

        :param extra_properties: A dictionary of extra properties to add to the event.
        :param extra_properties_func: A function that takes the function's arguments and returns
        a dictionary of extra properties to add to the event.
        
        :param include_duration: Whether to include the duration of the function in the event.
        :param include_results: Whether to include the results of the function in the event.
        :param include_args_in_properties: Whether to include the arguments of the function in
        the properties of the event.
        :param include_kwargs_in_properties: Whether to include the keyword arguments of the function in
        the properties of the event.

        :param usage_callback: A function that is the final step of the event logging. 
        It should take all the arguments of the function and results of the function and
        return a dictionary of usage properties to log return to the event.
        It should take the following keyword arguments:
            - result: Any The result of the function.
            - properties: Dict The properties of the event.
            - args: List The arguments of the function.
            - kwargs: Dict The keyword arguments of the function.
            - wrapper_kwargs: Dict The keyword arguments of the decorator.

        """

        wrapper_kwargs = kwargs or {}
        def decorator(func):
            if is_coro_func(func):
                @functools.wraps(func)
                async def wrapper(*args, **kwargs):

                    # validate exclude_on
                    if exclude_on is not None:
                        if is_coro_func(exclude_on) and await exclude_on(*args, **kwargs, **wrapper_kwargs):
                            return await func(*args, **kwargs)
                        elif exclude_on(*args, **kwargs):
                            return await func(*args, **kwargs)

                    # validate customer_id
                    if customer_id is None:
                        logger.error("customer_id is required for event logging.")
                        return await func(*args, **kwargs)
                    
                    _customer_id = customer_id
                    if callable(customer_id):
                        if is_coro_func(customer_id):
                            _customer_id = await customer_id(*args, **kwargs, **wrapper_kwargs)
                        else:
                            _customer_id = customer_id(*args, **kwargs, **wrapper_kwargs)
                    
                    # validate metric_id
                    if metric_id is None:
                        _metric_id = full_name(func)
                    elif callable(metric_id):
                        if is_coro_func(metric_id):
                            _metric_id = await metric_id(*args, **kwargs, **wrapper_kwargs)
                        else:
                            _metric_id = metric_id(*args, **kwargs, **wrapper_kwargs)
                    
                    
                    # validate/handle properties
                    event_properties = {
                        'operation_type': 'add'
                    }
                    if properties is not None:
                        if callable(properties):
                            if is_coro_func(properties):
                                event_properties = await properties(*args, **kwargs, **wrapper_kwargs)
                            else:
                                event_properties = properties(*args, **kwargs, **wrapper_kwargs)
                        else:
                            event_properties = properties
                    
                    if extra_properties:
                        event_properties.update(extra_properties)
                    
                    if extra_properties_func:
                        if is_coro_func(extra_properties_func):
                            event_properties.update(await extra_properties_func(*args, **kwargs, **wrapper_kwargs))
                        else:
                            event_properties.update(extra_properties_func(*args, **kwargs, **wrapper_kwargs))
                    
                    if exclude_properties:
                        for prop in exclude_properties:
                            event_properties.pop(prop, None)
                    
                    duration_s = timer() if include_duration else None
                    result = await func(*args, **kwargs)
                    if include_duration: event_properties['duration_s'] = timer(duration_s)
                    if include_results: event_properties['results'] = result

                    # Handle including args and kwargs in properties
                    if include_args_in_properties:
                        event_properties['args'] = args
                    if include_kwargs_in_properties:
                        event_properties['kwargs'] = kwargs
                    
                    # validate/handle transaction_id
                    if transaction_id is None:
                        _transaction_id = str(uuid.uuid4())
                    elif callable(transaction_id):
                        if is_coro_func(transaction_id):
                            _transaction_id = await transaction_id(*args, **kwargs, **wrapper_kwargs)
                        else:
                            _transaction_id = transaction_id(*args, **kwargs, **wrapper_kwargs)
                    
                    # validate/handle usage_callback
                    if usage_callback is not None:
                        if is_coro_func(usage_callback):
                            event_properties = await usage_callback(args = args, result = result, properties = event_properties, kwargs = kwargs, wrapper_kwargs = wrapper_kwargs)
                        else:
                            event_properties = usage_callback(args = args, result = result, properties = event_properties, kwargs = kwargs, wrapper_kwargs = wrapper_kwargs)

                    try:
                        await self.events.async_create(
                            transaction_id = _transaction_id,
                            external_customer_id = _customer_id,
                            code = _metric_id,
                            properties = event_properties,
                            timestamp = timestamp,
                            **kwargs,
                        )
                    except Exception as e:
                        logger.error(f"Error logging event: {e}")
                    return result
            
            else:
                @functools.wraps(func)
                def wrapper(*args, **kwargs):
                    # validate exclude_on
                    if exclude_on is not None and exclude_on(*args, **kwargs, **wrapper_kwargs):
                        return func(*args, **kwargs)

                    # validate customer_id
                    if customer_id is None:
                        logger.error("customer_id is required for event logging.")
                        return func(*args, **kwargs)
                    
                    _customer_id = customer_id
                    if callable(customer_id):
                        _customer_id = customer_id(*args, **kwargs, **wrapper_kwargs)
                    
                    # validate event_name
                    if metric_id is None:
                        _metric_id = full_name(func)
                    elif callable(metric_id):
                        _metric_id = metric_id(*args, **kwargs, **wrapper_kwargs)
                    
                    
                    # validate/handle properties
                    event_properties = {
                        'operation_type': 'add'
                    }
                    if properties is not None:
                        if callable(properties):
                            event_properties = properties(*args, **kwargs, **wrapper_kwargs)
                        else:
                            event_properties = properties
                    
                    if extra_properties:
                        event_properties.update(extra_properties)
                    
                    if extra_properties_func:
                        event_properties.update(extra_properties_func(*args, **kwargs, **wrapper_kwargs))
                    
                    if exclude_properties:
                        for prop in exclude_properties:
                            event_properties.pop(prop, None)
                    
                    duration_s = timer() if include_duration else None
                    result = func(*args, **kwargs)
                    
                    if include_duration: event_properties['duration_s'] = timer(duration_s)
                    if include_results: event_properties['results'] = result
                    
                    # Handle including args and kwargs in properties
                    if include_args_in_properties:
                        event_properties['args'] = args
                    if include_kwargs_in_properties:
                        event_properties['kwargs'] = kwargs

                    # validate/handle transaction_id
                    if transaction_id is None:
                        _transaction_id = str(uuid.uuid4())
                    elif callable(transaction_id):
                        _transaction_id = transaction_id(*args, **kwargs, **wrapper_kwargs)
                    
                    # validate/handle usage_callback
                    if usage_callback is not None:
                        event_properties = usage_callback(args = args, result = result, properties = event_properties, kwargs = kwargs, wrapper_kwargs = wrapper_kwargs)


                    try:
                        self.events.create(
                            transaction_id = _transaction_id,
                            external_customer_id = _customer_id,
                            code = _metric_id,
                            properties = event_properties,
                            timestamp = timestamp,
                            **kwargs,
                        )
                    except Exception as e:
                        logger.error(f"Error logging metric: {e}")

                    return result
                
            return wrapper
        return decorator



class LagoAPI:

    settings: Optional[LagoSettings] = lago_settings
    _api: Optional[LagoClient] = None

    """
    The Global Class for Lago API.
    """

    def configure(
        self, 
        api_key: Optional[str] = None,
        url: Optional[str] = None,
        scheme: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
        api_path: Optional[str] = None,

        debug_enabled: Optional[bool] = None,
        on_error: Optional[Callable] = None,
        timeout: Optional[int] = None,
        api_key_header: Optional[str] = None,
        ignore_errors: Optional[bool] = None,

        reset: Optional[bool] = True,
        **kwargs
    ):
        """
        Configure the global Lago client.
        """
        self.settings.configure(
            api_key=api_key,
            url=url,
            scheme=scheme,
            host=host,
            port=port,
            api_path=api_path,
            debug_enabled=debug_enabled,
            on_error=on_error,
            timeout=timeout,
            api_key_header=api_key_header,
            ignore_errors=ignore_errors,
            **kwargs
        )

        if reset: self._api = None
        if self._api is None:
            self.get_api(**kwargs)
    
    def get_api(self, **kwargs) -> LagoClient:
        if self._api is None:
            self._api = LagoClient(settings = self.settings, **kwargs)
        return self._api
    
    @property
    def api(self) -> LagoClient:
        """
        Returns the inherited Lago client.
        """
        if self._api is None:
            self.configure()
        return self._api
    

    @property
    def add_ons(self) -> AddOnRoute:
        """
        Returns the `AddOnRoute` class for interacting with add-ons.
        
        Doc: `https://doc.getlago.com/docs/api/add_ons/add-on-object`
        """
        return self.api.add_ons
    

    @property
    def applied_add_ons(self) -> AppliedAddOnRoute:
        """
        Returns the `AppliedAddOnRoute` class for interacting with applied add-ons.
        
        Doc: `https://doc.getlago.com/docs/api/add_ons/apply-add-on`
        """
        return self.api.applied_add_ons
    

    @property
    def applied_coupons(self) -> AppliedCouponRoute:
        """
        Returns the `AppliedCouponRoute` class for interacting with applied coupons.

        Doc: `https://doc.getlago.com/docs/api/coupons/coupon-object`
        """
        return self.api.applied_coupons
    

    @property
    def billable_metrics(self) -> BillableMetricRoute:
        """
        Returns the `BillableMetricRoute` class for interacting with billable metrics.

        Doc: `https://doc.getlago.com/docs/api/billable_metrics/billable-metric-object`
        """
        return self.api.billable_metrics
    

    @property
    def coupons(self) -> CouponRoute:
        """
        Returns the `CouponRoute` class for interacting with coupons.

        Doc: `https://doc.getlago.com/docs/api/coupons/coupon-object`
        """
        return self.api.coupons
    

    @property
    def customers(self) -> CustomerRoute:
        """
        Returns the `CustomerRoute` class for interacting with customers.

        Doc: `https://doc.getlago.com/docs/api/customers/customer-object`
        """
        return self.api.customers
    

    @property
    def events(self) -> EventRoute:
        """
        Returns the `EventRoute` class for interacting with events.

        Doc: `https://doc.getlago.com/docs/api/events/event-object`
        """
        return self.api.events
    

    @property
    def invoices(self) -> InvoiceRoute:
        """
        Returns the `InvoiceRoute` class for interacting with invoices.

        Doc: `https://doc.getlago.com/docs/api/invoices/invoice-object`
        """
        return self.api.invoices
    

    @property
    def organizations(self) -> OrganizationRoute:
        """
        Returns the `OrganizationRoute` class for interacting with organizations.

        Doc: `https://doc.getlago.com/docs/api/organizations/organization-object`
        """
        return self.api.organizations
    

    @property
    def plans(self) -> PlanRoute:
        """
        Returns the `PlanRoute` class for interacting with plans.

        Doc: `https://doc.getlago.com/docs/api/plans/plan-object`
        """
        return self.api.plans
    

    @property
    def subscriptions(self) -> SubscriptionRoute:
        """
        Returns the `SubscriptionRoute` class for interacting with subscriptions.

        Doc: `https://doc.getlago.com/docs/api/subscriptions/subscription-object`
        """
        return self.api.subscriptions
    

    @property
    def wallets(self) -> WalletRoute:
        """
        Returns the `WalletRoute` class for interacting with wallets.

        Doc: `https://doc.getlago.com/docs/api/wallets/wallet-object`
        """
        return self.api.wallets
    

    @property
    def wallet_transactions(self) -> WalletTransactionRoute:
        """
        Returns the `WalletTransactionRoute` class for interacting with wallet transactions.

        Doc: `https://doc.getlago.com/docs/api/wallets/wallet-transaction-object`
        """
        return self.api.wallet_transactions
    

    @property
    def webhooks(self) -> WebhookRoute:
        """
        Returns the `WebhookRoute` class for interacting with webhooks.

        Doc: `https://doc.getlago.com/docs/api/webhooks/webhook-object`
        """
        return self.api.webhooks
    
    def on_event(
        self,
        customer_id: Optional[Union[str, Callable]] = None,
        metric_id: Optional[Union[str, Callable]] = None,
        transaction_id: Optional[Union[str, Callable]] = None,
        
        properties: Optional[Union[Dict, Callable]] = None,
        created_at: Optional[datetime.datetime] = None,
        exclude_properties: Optional[List[str]] = None,
        exclude_on: Optional[Callable] = None,
        extra_properties: Optional[Dict] = None,
        extra_properties_func: Optional[Callable] = None,
        include_duration: Optional[bool] = False,
        include_results: Optional[bool] = True,
        include_args_in_properties: Optional[bool] = False,
        include_kwargs_in_properties: Optional[bool] = False,

        usage_callback: Optional[Callable] = None,
        **kwargs,
    ):
        """
        Lago Event Decorator to log events to the Lago API.
        Most params are either _str_ or _Callable_ which will pass the function's arguments 
        and expects a _str_ to be returned. If required parameters are not provided, the 
        event logging will not occur.

        Example:

        >>> @Lago.on_event(customer_id = get_customer_id)
        ... def my_func(headers: Dict, params: Dict):
        ...     ...

        >>> @Lago.on_event(customer_id="1234")
        ... async def async_run_func():
        ...     ...

        >>> @Lago.on_event(customer_id="1234", metric_id="my_event")
        ... def run_func():
        ...     ...

        :param customer_id: The customer_id to log the event to. (required)
        :param metric_id: The `metric_id` which correlates to a `BillableMetrics` to log. [also known as `code`]. defaults to the function name.
        :param transaction_id: Unique identifer of the event in your application. Defaults to a unique uuid.

        :param properties: A dictionary of parameters to log with the event. Defaults 
        to the function's arguments.
        :param created_at: The time the event occurred. Defaults to the current time.

        :param exclude_properties: A list of properties to exclude from the event.
        :param exclude_on: A function that takes validates whether this event should be logged.

        :param extra_properties: A dictionary of extra properties to add to the event.
        :param extra_properties_func: A function that takes the function's arguments and returns
        a dictionary of extra properties to add to the event.
        
        :param include_duration: Whether to include the duration of the function in the event.
        :param include_results: Whether to include the results of the function in the event.
        :param include_args_in_properties: Whether to include the arguments of the function in
        the properties of the event.
        :param include_kwargs_in_properties: Whether to include the keyword arguments of the function in
        the properties of the event.
        :param usage_callback: A function that is the final step of the event logging. 
        It should take all the arguments of the function and results of the function and
        return a dictionary of usage properties to log return to the event.
        It should take the following keyword arguments:
            - result: Any The result of the function.
            - properties: Dict The properties of the event.
            - args: List The arguments of the function.
            - kwargs: Dict The keyword arguments of the function.
            - wrapper_kwargs: Dict The keyword arguments of the decorator.


        """
        return self.api.on_event(
            customer_id=customer_id,
            metric_id=metric_id,
            transaction_id=transaction_id,
            properties=properties,
            created_at=created_at,
            exclude_properties=exclude_properties,
            exclude_on=exclude_on,
            extra_properties=extra_properties,
            extra_properties_func=extra_properties_func,
            include_duration=include_duration,
            include_results=include_results,
            include_args_in_properties=include_args_in_properties,
            include_kwargs_in_properties=include_kwargs_in_properties,
            usage_callback=usage_callback,
            **kwargs,
        )

    """
    Context Managers
    """

    async def async_close(self):
        if self._api is not None:
            await self._api.async_close()
    
    def close(self):
        if self._api is not None:
            self._api.close()
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        self.close()
    
    async def __aenter__(self):
        return self
    
    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.async_close()


    
Lago: LagoAPI = LagoAPI()










from typing import Optional, Type
from aiolago.types.base import BaseRoute, BaseResource, lazyproperty

__all__ = [
    'Wallet',
    'WalletResponse',
    'WalletRoute',
]

class Wallet(BaseResource):
    name: Optional[str]
    external_customer_id: Optional[str]
    rate_amount: Optional[str] = '1.0'
    currency: Optional[str] = 'USD'
    paid_credits: Optional[str]
    granted_credits: Optional[str]
    expiration_date: Optional[str]


class WalletResponse(BaseResource):
    lago_id: str
    lago_customer_id: str
    external_customer_id: str
    status: str
    currency: str
    name: Optional[str]
    rate_amount: str
    credits_balance: str
    balance: str
    consumed_credits: str
    created_at: str
    expiration_date: Optional[str]
    last_balance_sync_at: Optional[str]
    last_consumed_credit_at: Optional[str]
    terminated_at: Optional[str]

class WalletRoute(BaseRoute):
    input_model: Optional[Type[BaseResource]] = Wallet
    response_model: Optional[Type[BaseResource]] = WalletResponse
    usage_model: Optional[Type[BaseResource]] = None

    @lazyproperty
    def api_resource(self):
        return 'wallets'

    @lazyproperty
    def root_name(self):
        return 'wallet'


from typing import Optional, Type
from aiolago.types.base import BaseRoute, BaseResource, lazyproperty

__all__ = [
    'AddOn',
    'AddOnResponse',
    'AddOnRoute',
]

class AddOn(BaseResource):
    name: Optional[str]
    code: Optional[str]
    amount_cents: Optional[int]
    amount_currency: Optional[str]
    description: Optional[str]


class AddOnResponse(BaseResource):
    lago_id: str
    name: str
    code: str
    amount_cents: int
    amount_currency: str
    created_at: str
    description: Optional[str]


class AddOnRoute(BaseRoute):
    input_model: Optional[Type[BaseResource]] = AddOn
    response_model: Optional[Type[BaseResource]] = AddOnResponse
    usage_model: Optional[Type[BaseResource]] = None

    @lazyproperty
    def api_resource(self):
        return 'add_ons'

    @lazyproperty
    def root_name(self):
        return 'add_on'
    


import uuid
from typing import Optional, Type, List
from aiolago.types.base import BaseRoute, BaseResource, lazyproperty, Field

__all__ = [
    'BillableMetric',
    'BillableMetricResponse',
    'BillableMetricRoute',
]

def transaction_id():
    return str(uuid.uuid4())

class Event(BaseResource):
    transaction_id: str = Field(default_factory = transaction_id)
    external_customer_id: Optional[str]
    external_subscription_id: Optional[str]
    code: str
    timestamp: Optional[int]
    properties: Optional[dict]


class BatchEvent(BaseResource):
    transaction_id: str = Field(default_factory = transaction_id)
    external_customer_id: Optional[str]
    external_subscription_ids: List[str]
    code: str
    timestamp: Optional[int]
    properties: Optional[dict]


class EventResponse(BaseResource):
    lago_id: str
    transaction_id: str
    external_customer_id: Optional[str]
    lago_customer_id: Optional[str]
    lago_subscription_id: Optional[str]
    external_subscription_id: Optional[str]
    code: str
    timestamp: str
    properties: Optional[dict]
    created_at: str


class EventRoute(BaseRoute):
    input_model: Optional[Type[BaseResource]] = Event
    response_model: Optional[Type[BaseResource]] = EventResponse
    usage_model: Optional[Type[BaseResource]] = None
    
    @lazyproperty
    def api_resource(self):
        return 'events'
    
    @lazyproperty
    def root_name(self):
        return 'event'
    
    def batch_create(
        self, 
        input_object: BatchEvent,
        return_response: bool = False,
        **kwargs
    ):
        """
        Batch Create Resources

        :param input_object: Input Object to Create
        :param return_response: Return the Response Object
        """
        return super().batch_create(
            input_object=input_object,
            return_response=return_response,
            **kwargs
        )
    
    async def async_batch_create(
        self, 
        input_object: BatchEvent,
        return_response: bool = False,
        **kwargs
    ):
        """
        Batch Create Resources

        :param input_object: Input Object to Create
        :param return_response: Return the Response Object
        """
        return await super().async_batch_create(
            input_object=input_object,
            return_response=return_response,
            **kwargs
        )




from typing import Optional, Type, List, Union, Dict, Any
from aiolago.types.base import BaseRoute, BaseResource, lazyproperty

__all__ = [
    'Charge',
    'Charges',
    'Plan',
    'PlanResponse',
    'PlanRoute',
]


class Charge(BaseResource):
    id: Optional[str]
    lago_id: Optional[str]
    billable_metric_id: Optional[str]
    lago_billable_metric_id: Optional[str]
    charge_model: Optional[str]
    properties: Optional[Union[dict, list]]
    group_properties: Optional[Union[List[Any], Dict]]

class Charges(BaseResource):
    __root__: List[Charge] = []

class Plan(BaseResource):
    name: Optional[str]
    code: Optional[str]
    interval: Optional[str]
    description: Optional[str]
    amount_cents: Optional[int]
    amount_currency: Optional[str]
    trial_period: Optional[float]
    pay_in_advance: Optional[bool]
    bill_charges_monthly: Optional[bool]
    charges: Optional[Union[List[Charge], Charges]] = Charges()
    
    def add_charge_to_plan(self, charge: Charge):
        """
        Adds a charge to the plan's charges.
        """
        self.charges.__root__.append(charge)


class PlanResponse(BaseResource):
    lago_id: str
    name: str
    created_at: str
    code: str
    interval: Optional[str]
    description: Optional[str]
    amount_cents: Optional[int]
    amount_currency: Optional[str]
    trial_period: Optional[float]
    pay_in_advance: Optional[bool]
    bill_charges_monthly: Optional[bool]
    charges: Optional[Charges]

    def has_metric_id(self, metric_id: str) -> bool:
        """
        Checks to see whether a metric_id is present in the plan's charges.
        """
        return any(
            (charge.billable_metric_id and charge.billable_metric_id == metric_id) or 
            (charge.lago_billable_metric_id and charge.lago_billable_metric_id == metric_id) 
            for charge in self.charges.__root__
        )
    
    def remove_charge_from_plan(self, charge: Charge):
        """
        Removes a charge from the plan's charges.
        """
        for c in self.charges.__root__:
            if (c.billable_metric_id and c.billable_metric_id == charge.billable_metric_id) or (c.lago_billable_metric_id and c.lago_billable_metric_id == charge.lago_billable_metric_id):
                self.charges.__root__.remove(c)
                break

    
    def to_plan(self) -> Plan:
        """
        Converts a PlanResponse to a Plan.
        """
        return Plan(
            name = self.name,
            code = self.code,
            interval = self.interval,
            description = self.description,
            amount_cents = self.amount_cents,
            amount_currency = self.amount_currency,
            trial_period = self.trial_period,
            pay_in_advance = self.pay_in_advance,
            bill_charges_monthly = self.bill_charges_monthly,
            charges = self.charges
        )

    def add_charge_to_plan(self, charge: Charge):
        """
        Adds a charge to the plan's charges.
        """
        self.charges.__root__.append(charge)

class PlanRoute(BaseRoute):
    input_model: Optional[Type[BaseResource]] = Plan
    response_model: Optional[Type[BaseResource]] = PlanResponse
    usage_model: Optional[Type[BaseResource]] = None

    @lazyproperty
    def api_resource(self):
        return 'plans'

    @lazyproperty
    def root_name(self):
        return 'plan'


from typing import Optional, Dict
from lazyops.types import BaseSettings, lazyproperty
from aiolago.version import VERSION



class LagoSettings(BaseSettings):

    url: Optional[str] = None
    scheme: Optional[str] = 'http://'
    host: Optional[str] = None
    port: Optional[int] = 3000
    
    api_key: Optional[str] = None
    api_path: Optional[str] = '/api/v1/'

    api_key_header: Optional[str] = None
    organization_id: Optional[str] = None

    ignore_errors: Optional[bool] = False
    debug_enabled: Optional[bool] = True

    timeout: Optional[int] = 10
    max_retries: Optional[int] = 3
    
    class Config:
        env_prefix = 'LAGO_'
        case_sensitive = False

    @lazyproperty
    def api_url(self) -> str:
        if self.url:
            return self.url
        if self.host:
            url = f"{self.scheme}{self.host}"
            if self.port: url += f":{self.port}"
            return url
        
        # Return the official Lago API URL
        return "https://api.getlago.com"
    
    @lazyproperty
    def base_url(self) -> str:
        if self.api_path:
            from urllib.parse import urljoin
            return urljoin(self.api_url, self.api_path)
        return self.api_url
    
    @lazyproperty
    def headers(self):
        _headers = {"Content-Type": "application/json", "User-Agent": f"aiolago/{VERSION}"}
        if self.api_key: 
            if self.api_key_header:
                _headers[self.api_key_header] = self.api_key
            else:
                _headers['Authorization'] = f'Bearer {self.api_key}'
        return _headers


    def get_headers(
        self, 
        api_key: Optional[str] = None, 
        api_key_header: Optional[str] = None, 
        **kwargs
    ) -> Dict[str, str]:

        headers = self.headers.copy()
        if kwargs: headers.update(**kwargs)
        api_key_header = api_key_header if api_key_header is not None else self.api_key_header
        api_key = api_key if api_key is not None else self.api_key
        if api_key:
            if api_key_header:
                headers[api_key_header] = api_key
            else:
                headers['Authorization'] = f'Bearer {api_key}'
        return headers

    def get_api_url(
        self, 
        host: Optional[str] = None, 
        port: Optional[int] = None, 
        scheme: Optional[str] = None, 
        url: Optional[str] = None,
        **kwargs
    ) -> str:
        if url: return url
        if host:
            url = f"{scheme or self.scheme}{host}"
            if port: url += f":{port}"
            return url
        return self.api_url

    def get_base_api_url(
        self, 
        host: Optional[str] = None, 
        port: Optional[int] = None, 
        scheme: Optional[str] = None, 
        url: Optional[str] = None,
        api_path: Optional[str] = None,
        **kwargs
    ) -> str:
        api_url = self.get_api_url(
            host=host,
            port=port,
            scheme=scheme,
            url=url,
        )
        api_path = api_path or self.api_path
        if api_path:
            from urllib.parse import urljoin
            return urljoin(api_url, api_path)
        return api_url
    
    def configure(
        self,
        url: Optional[str] = None,
        scheme: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
        api_key: Optional[str] = None,
        api_path: Optional[str] = None,
        api_key_header: Optional[str] = None,
        organization_id: Optional[str] = None,
        ignore_errors: Optional[bool] = None,
        debug_enabled: Optional[bool] = None,
        timeout: Optional[int] = None,
        max_retries: Optional[int] = None,
        **kwargs,
    ):
        """
        Configure the Lago API client.

        :param url: The Lago API URL.
        :param scheme: The Lago API scheme.
        :param host: The Lago API host.
        :param port: The Lago API port.
        :param api_key: The Lago API key.
        :param api_path: The Lago API path.
        :param api_key_header: The Lago API key header.
        :param organization_id: The Lago organization ID.
        :param ignore_errors: Ignore Lago API errors.
        :param debug_enabled: Enable debug mode.
        :param timeout: Timeout in seconds.
        :param max_retries: Maximum number of retries.
        """
        if url is not None: self.url = url
        if scheme is not None: self.scheme = scheme
        if host is not None: self.host = host
        if port is not None: self.port = port
        if api_key is not None: self.api_key = api_key
        if api_path is not None: self.api_path = api_path
        if api_key_header is not None: self.api_key_header = api_key_header
        if organization_id is not None: self.organization_id = organization_id
        if ignore_errors is not None: self.ignore_errors = ignore_errors
        if debug_enabled is not None: self.debug_enabled = debug_enabled
        if timeout is not None: self.timeout = timeout
        if max_retries is not None: self.max_retries = max_retries
        if kwargs:
            for key, value in kwargs.items():
                if hasattr(self, key):
                    setattr(self, key, value)



settings = LagoSettings()


from typing import Optional, Type
from aiolago.types.base import BaseRoute, BaseResource, lazyproperty

__all__ = [
    'WalletTransaction',
    'WalletTransactionResponse',
    'WalletTransactionRoute',
]


class WalletTransaction(BaseResource):
    wallet_id: Optional[str]
    paid_credits: Optional[str]
    granted_credits: Optional[str]


class WalletTransactionResponse(BaseResource):
    lago_id: str
    lago_wallet_id: str
    status: str
    transaction_type: str
    amount: str
    credit_amount: str
    settled_at: Optional[str]
    created_at: str


class WalletTransactionRoute(BaseRoute):
    input_model: Optional[Type[BaseResource]] = WalletTransaction
    response_model: Optional[Type[BaseResource]] = WalletTransactionResponse
    usage_model: Optional[Type[BaseResource]] = None

    @lazyproperty
    def api_resource(self):
        return 'wallet_transactions'

    @lazyproperty
    def root_name(self):
        return 'wallet_transaction'

